@@ -1,5 +1,6 @@
-from typing import Callable, List, Mapping
+from typing import Callable, List, Mapping, Protocol, Optional
 
+from debputy.commands.debputy_cmd.output import IOBasedOutputStyling
 from debputy.dh_migration.migrators_impl import (
     migrate_links_files,
     migrate_maintscript,
@@ -22,6 +23,7 @@ from debputy.dh_migration.migrators_impl import (
     detect_obsolete_substvars,
     detect_dh_addons_zz_debputy_rrr,
     detect_dh_addons_with_full_integration,
+    migrate_shell_completions,
 )
 from debputy.dh_migration.models import AcceptableMigrationIssues, FeatureMigration
 from debputy.highlevel_manifest import HighLevelManifest
@@ -61,6 +63,7 @@ _DH_DEBPUTY_MIGRATORS = [
     migrate_tmpfile,
     migrate_lintian_overrides_files,
     migrate_bash_completion,
+    migrate_shell_completions,
     detect_obsolete_substvars,
     # not-installed should go last, so its rules appear after other installations
     # It is not perfect, but it is a start.
