@@ -4,10 +4,10 @@ import stat
 import subprocess
 import sys
 import textwrap
-from typing import Optional, List, Union, NoReturn, Mapping
+from typing import Optional, List, Union, NoReturn, Mapping, Sequence, TYPE_CHECKING
 
 from debputy.commands.debputy_cmd.context import CommandContext
-from debputy.commands.debputy_cmd.output import _output_styling, OutputStylingBase
+from debputy.commands.debputy_cmd.output import _output_styling, IOBasedOutputStyling
 from debputy.filesystem_scan import FSROOverlay
 from debputy.linting.lint_util import (
     LintReport,
@@ -17,24 +17,29 @@ from debputy.linting.lint_util import (
     LintDiagnosticResultState,
     AsyncLinterImpl,
 )
+from debputy.lsp.config.debputy_config import DebputyConfig, load_debputy_config
 from debputy.lsp.diagnostics import DiagnosticData
-from debputy.lsp.lsp_debian_changelog import _lint_debian_changelog
-from debputy.lsp.lsp_debian_control import (
+from debputy.lsp.languages.lsp_debian_changelog import _lint_debian_changelog
+from debputy.lsp.languages.lsp_debian_control import (
     _lint_debian_control,
     _reformat_debian_control,
 )
-from debputy.lsp.lsp_debian_copyright import (
+from debputy.lsp.languages.lsp_debian_copyright import (
     _lint_debian_copyright,
     _reformat_debian_copyright,
 )
-from debputy.lsp.lsp_debian_debputy_manifest import _lint_debian_debputy_manifest
-from debputy.lsp.lsp_debian_patches_series import _lint_debian_patches_series
-from debputy.lsp.lsp_debian_rules import _lint_debian_rules
-from debputy.lsp.lsp_debian_tests_control import (
+from debputy.lsp.languages.lsp_debian_debputy_manifest import (
+    _lint_debian_debputy_manifest,
+)
+from debputy.lsp.languages.lsp_debian_patches_series import _lint_debian_patches_series
+from debputy.lsp.languages.lsp_debian_rules import _lint_debian_rules
+from debputy.lsp.languages.lsp_debian_tests_control import (
     _lint_debian_tests_control,
     _reformat_debian_tests_control,
 )
-from debputy.lsp.lsp_debian_upstream_metadata import _lint_debian_upstream_metadata
+from debputy.lsp.languages.lsp_debian_upstream_metadata import (
+    _lint_debian_upstream_metadata,
+)
 from debputy.lsp.maint_prefs import (
     MaintainerPreferenceTable,
     EffectiveFormattingPreference,
@@ -49,18 +54,6 @@ from debputy.lsp.text_edit import (
     OverLappingTextEditException,
 )
 from debputy.lsp.vendoring._deb822_repro import Deb822FileElement
-from debputy.lsprotocol.types import (
-    CodeAction,
-    Command,
-    CodeActionParams,
-    CodeActionContext,
-    TextDocumentIdentifier,
-    TextEdit,
-    Position,
-    Range,
-    DiagnosticSeverity,
-    Diagnostic,
-)
 from debputy.packages import SourcePackage, BinaryPackage
 from debputy.plugin.api import VirtualPath
 from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
@@ -68,6 +61,12 @@ from debputy.util import _warn, _error, _info
 from debputy.yaml import MANIFEST_YAML, YAMLError
 from debputy.yaml.compat import CommentedMap
 
+if TYPE_CHECKING:
+    import lsprotocol.types as types
+else:
+    import debputy.lsprotocol.types as types
+
+
 LINTER_FORMATS = {
     "debian/changelog": _lint_debian_changelog,
     "debian/control": _lint_debian_control,
@@ -93,6 +92,7 @@ class LintContext:
     maint_preference_table: MaintainerPreferenceTable
     source_root: Optional[VirtualPath]
     debian_dir: Optional[VirtualPath]
+    debputy_config: DebputyConfig
     parsed_deb822_file_content: Optional[Deb822FileElement] = None
     source_package: Optional[SourcePackage] = None
     binary_packages: Optional[Mapping[str, BinaryPackage]] = None
@@ -116,6 +116,7 @@ class LintContext:
             path,
             content,
             lines,
+            self.debputy_config,
             self.source_package,
             self.binary_packages,
             self.effective_preference,
@@ -133,6 +134,7 @@ def gather_lint_info(context: CommandContext) -> LintContext:
         MaintainerPreferenceTable.load_preferences(),
         source_root,
         debian_dir,
+        load_debputy_config(),
     )
     try:
         with open("debian/control") as fd:
@@ -349,7 +351,7 @@ def perform_reformat(
 
 
 def perform_reformat_of_file(
-    fo: OutputStylingBase,
+    fo: IOBasedOutputStyling,
     lint_context: LintContext,
     filename: str,
     formatter: FormatterImpl,
@@ -403,8 +405,8 @@ def perform_reformat_of_file(
 def _exit_with_lint_code(lint_report: LintReport) -> NoReturn:
     diagnostics_count = lint_report.diagnostics_count
     if (
-        diagnostics_count[DiagnosticSeverity.Error]
-        or diagnostics_count[DiagnosticSeverity.Warning]
+        diagnostics_count[types.DiagnosticSeverity.Error]
+        or diagnostics_count[types.DiagnosticSeverity.Warning]
     ):
         sys.exit(2)
     sys.exit(0)
@@ -442,8 +444,8 @@ async def perform_linting_of_file(
 
 
 def _overlapping_edit(
-    last_edit_range: Range,
-    last_fix_range: Range,
+    last_edit_range: types.Range,
+    last_fix_range: types.Range,
 ) -> bool:
     last_edit_start_pos = last_edit_range.start
     last_fix_start_pos = last_fix_range.start
@@ -464,9 +466,9 @@ def _overlapping_edit(
 
 
 def _max_range(
-    r1: Range,
-    r2: Range,
-) -> Range:
+    r1: types.Range,
+    r2: types.Range,
+) -> types.Range:
     if r2.end > r1.end:
         return r2
     if r2.end < r1.end:
@@ -476,13 +478,13 @@ def _max_range(
     return r1
 
 
-_INITIAL_FIX_RANGE = Range(
-    Position(0, 0),
-    Position(0, 0),
+_INITIAL_FIX_RANGE = types.Range(
+    types.Position(0, 0),
+    types.Position(0, 0),
 )
 
 
-def _is_non_interactive_auto_fix_allowed(diagnostic: Diagnostic) -> bool:
+def _is_non_interactive_auto_fix_allowed(diagnostic: types.Diagnostic) -> bool:
     diag_data: Optional[DiagnosticData] = diagnostic.data
     return diag_data is None or diag_data.get("enable_non_interactive_auto_fix", True)
 
@@ -495,7 +497,7 @@ async def _auto_fix_run(
     lint_report: LintReport,
 ) -> None:
     another_round = True
-    unfixed_diagnostics: List[Diagnostic] = []
+    unfixed_diagnostics: List[types.Diagnostic] = []
     remaining_rounds = 10
     fixed_count = 0
     too_many_rounds = False
@@ -519,15 +521,16 @@ async def _auto_fix_run(
                 unfixed_diagnostics.append(diagnostic)
                 continue
             actions = provide_standard_quickfixes_from_diagnostics_lint(
-                CodeActionParams(
-                    TextDocumentIdentifier(filename),
+                lint_state,
+                types.CodeActionParams(
+                    types.TextDocumentIdentifier(lint_state.doc_uri),
                     diagnostic.range,
-                    CodeActionContext(
+                    types.CodeActionContext(
                         [diagnostic],
                     ),
                 ),
             )
-            auto_fixing_edits = resolve_auto_fixer(filename, actions)
+            auto_fixing_edits = resolve_auto_fixer(lint_state.doc_uri, actions)
 
             if not auto_fixing_edits:
                 unfixed_diagnostics.append(diagnostic)
@@ -665,15 +668,16 @@ async def _diagnostics_run(
         issues = await lint_state.gather_diagnostics()
         for diagnostic in issues:
             actions = provide_standard_quickfixes_from_diagnostics_lint(
-                CodeActionParams(
-                    TextDocumentIdentifier(filename),
+                lint_state,
+                types.CodeActionParams(
+                    types.TextDocumentIdentifier(lint_state.doc_uri),
                     diagnostic.range,
-                    CodeActionContext(
+                    types.CodeActionContext(
                         [diagnostic],
                     ),
                 ),
             )
-            auto_fixer = resolve_auto_fixer(filename, actions)
+            auto_fixer = resolve_auto_fixer(lint_state.doc_uri, actions)
             has_auto_fixer = bool(auto_fixer) and _is_non_interactive_auto_fix_allowed(
                 diagnostic
             )
@@ -688,12 +692,15 @@ async def _diagnostics_run(
 
 
 def has_at_least_lsp_quickfix(
-    actions: Optional[List[Union[Command, CodeAction]]],
+    actions: Optional[List[Union[types.Command, types.CodeAction]]],
 ) -> bool:
     if actions is None:
         return False
     for action in actions:
-        if not isinstance(action, CodeAction):
+        if (
+            not isinstance(action, types.CodeAction)
+            or action.kind != types.CodeActionKind.QuickFix
+        ):
             continue
         if action.edit is None and action.command is None:
             continue
@@ -703,20 +710,30 @@ def has_at_least_lsp_quickfix(
 
 def resolve_auto_fixer(
     document_ref: str,
-    actions: Optional[List[Union[Command, CodeAction]]],
-) -> Optional[List[TextEdit]]:
+    actions: Optional[List[Union[types.Command, types.CodeAction]]],
+) -> Sequence[Union[types.TextEdit, types.AnnotatedTextEdit]]:
     if actions is None or len(actions) != 1:
-        return None
+        return tuple()
     action = actions[0]
-    if not isinstance(action, CodeAction):
-        return None
+    if not isinstance(action, types.CodeAction):
+        return tuple()
     workspace_edit = action.edit
     if workspace_edit is None or action.command is not None:
-        return None
+        return tuple()
+    document_changes = workspace_edit.document_changes
+    if document_changes:
+        if len(document_changes) != 1:
+            return tuple()
+        doc_edit = document_changes[0]
+        if not isinstance(doc_edit, types.TextDocumentEdit):
+            return tuple()
+        if doc_edit.text_document.uri != document_ref:
+            return tuple()
+        return doc_edit.edits
     if (
         not workspace_edit.changes
         or len(workspace_edit.changes) != 1
         or document_ref not in workspace_edit.changes
     ):
-        return None
+        return tuple()
     return workspace_edit.changes[document_ref]
