@@ -1,8 +1,10 @@
 import asyncio
+import collections
 import dataclasses
 import os
 import time
 from collections.abc import AsyncIterable
+from contextlib import suppress
 from typing import (
     Optional,
     List,
@@ -15,9 +17,16 @@ from typing import (
     Set,
     Dict,
     Iterable,
+    Sequence,
+    Callable,
+    Type,
 )
 
 import debputy.l10n as l10n
+from debputy.commands.debputy_cmd.output import (
+    OutputStyle,
+    MarkdownOutputStyle,
+)
 from debputy.dh.dh_assistant import (
     parse_drules_for_addons,
     DhSequencerData,
@@ -29,13 +38,15 @@ from debputy.linting.lint_util import (
     LintState,
     AsyncLinterImpl,
     AbortTaskError,
+    WorkspaceTextEditSupport,
 )
 from debputy.lsp.apt_cache import AptCache
+from debputy.lsp.config.debputy_config import load_debputy_config, DebputyConfig
 from debputy.lsp.diagnostics import DiagnosticReport
 from debputy.lsp.maint_prefs import (
     MaintainerPreferenceTable,
-    MaintainerPreference,
     determine_effective_preference,
+    EffectiveFormattingPreference,
 )
 from debputy.lsp.text_util import LintCapablePositionCodec
 from debputy.lsp.vendoring._deb822_repro import Deb822FileElement, parse_deb822_file
@@ -53,7 +64,7 @@ if TYPE_CHECKING:
     import lsprotocol.types as types
     from pygls.server import LanguageServer
     from pygls.workspace import TextDocument
-    from pygls.uris import from_fs_path
+    from pygls.uris import from_fs_path, to_fs_path
 
 else:
     import debputy.lsprotocol.types as types
@@ -61,7 +72,7 @@ else:
     try:
         from pygls.server import LanguageServer
         from pygls.workspace import TextDocument
-        from pygls.uris import from_fs_path
+        from pygls.uris import from_fs_path, to_fs_path
     except ImportError as e:
 
         class LanguageServer:
@@ -97,16 +108,17 @@ class FileCache:
         new_content: Optional[str] = None
         if is_open:
             last_doc_version = self.last_doc_version
-            dctrl_doc_version = doc.version
+            current_doc_version = doc.version
             if (
                 not self.is_open_in_editor
                 or last_doc_version is None
-                or dctrl_doc_version is None
-                or last_doc_version < dctrl_doc_version
+                or current_doc_version is None
+                or last_doc_version < current_doc_version
             ):
                 new_content = doc.source
 
             self.last_doc_version = doc.version
+            self.is_open_in_editor = True
         elif doc.uri.startswith("file://"):
             try:
                 with open(self.path) as fd:
@@ -120,6 +132,7 @@ class FileCache:
                 self._clear_cache()
                 self.is_valid = False
                 return False
+        self.is_open_in_editor = is_open
         if new_content is not None:
             self._update_cache(doc, new_content)
         self.is_valid = True
@@ -222,34 +235,43 @@ class LSProvidedLintState(LintState):
         dctrl_file = os.path.join(debian_dir_path, "control")
 
         if dctrl_file != doc.path:
-            self._dctrl_cache: DctrlFileCache = DctrlFileCache(
+
+            self._dctrl_cache: DctrlFileCache = ls.file_cache_for(
                 from_fs_path(dctrl_file),
                 dctrl_file,
-                dctrl_parser=dctrl_parser,
+                DctrlFileCache,
+                lambda uri, path: DctrlFileCache(uri, path, dctrl_parser=dctrl_parser),
             )
-            self._deb822_file: Deb822FileCache = Deb822FileCache(
+            self._deb822_file: Deb822FileCache = ls.file_cache_for(
                 doc.uri,
                 doc.path,
+                Deb822FileCache,
+                lambda uri, path: Deb822FileCache(uri, path),
             )
         else:
-            self._dctrl_cache: DctrlFileCache = DctrlFileCache(
+            self._dctrl_cache = ls.file_cache_for(
                 doc.uri,
                 doc.path,
-                dctrl_parser=dctrl_parser,
+                DctrlFileCache,
+                lambda uri, path: DctrlFileCache(uri, path, dctrl_parser=dctrl_parser),
             )
             self._deb822_file = self._dctrl_cache
 
         self._salsa_ci_caches = [
-            SalsaCICache(
+            ls.file_cache_for(
                 from_fs_path(os.path.join(debian_dir_path, p)),
                 os.path.join(debian_dir_path, p),
+                SalsaCICache,
+                lambda uri, path: SalsaCICache(uri, path),
             )
             for p in ("salsa-ci.yml", os.path.join("..", ".gitlab-ci.yml"))
         ]
         drules_path = os.path.join(debian_dir_path, "rules")
-        self._drules_cache = DebianRulesCache(
+        self._drules_cache = ls.file_cache_for(
             from_fs_path(drules_path) if doc.path != drules_path else doc.uri,
             drules_path,
+            DebianRulesCache,
+            lambda uri, path: DebianRulesCache(uri, path),
         )
 
     @property
@@ -260,6 +282,10 @@ class LSProvidedLintState(LintState):
     def doc_uri(self) -> str:
         return self._doc.uri
 
+    @property
+    def doc_version(self) -> Optional[int]:
+        return self._doc.version
+
     @property
     def source_root(self) -> Optional[VirtualPathBase]:
         return self._source_root
@@ -310,7 +336,7 @@ class LSProvidedLintState(LintState):
         return None
 
     @property
-    def effective_preference(self) -> Optional[MaintainerPreference]:
+    def effective_preference(self) -> Optional[EffectiveFormattingPreference]:
         source_package = self.source_package
         salsa_ci = self._resolve_salsa_ci()
         if source_package is None and salsa_ci is None:
@@ -348,6 +374,14 @@ class LSProvidedLintState(LintState):
             saw_dh,
         )
 
+    @property
+    def workspace_text_edit_support(self) -> WorkspaceTextEditSupport:
+        return self._ls.workspace_text_edit_support
+
+    @property
+    def debputy_config(self) -> DebputyConfig:
+        return self._ls.debputy_config
+
     def translation(self, domain: str) -> Translations:
         return self._ls.translation(domain)
 
@@ -442,6 +476,12 @@ class DebputyLanguageServer(LanguageServer):
         self.forced_locale: Optional[str] = None
         self._active_locale: Optional[List[str]] = None
         self._diagnostic_reports: Dict[str, DiagnosticReport] = {}
+        self.workspace_text_edit_support = WorkspaceTextEditSupport()
+        self.debputy_config = load_debputy_config()
+        self._file_state_caches: Dict[str, Dict[Type[FileCache], FileCache]] = (
+            collections.defaultdict(dict)
+        )
+        self.hover_output_style = OutputStyle()
 
     def finish_startup_initialization(self) -> None:
         if self._finished_initialization:
@@ -453,6 +493,13 @@ class DebputyLanguageServer(LanguageServer):
         _info(
             f"Loaded style preferences: {len(self.maint_preferences.maintainer_preferences)} unique maintainer preferences recorded"
         )
+        if (
+            self.hover_markup_format(
+                types.MarkupKind.Markdown, types.MarkupKind.PlainText
+            )
+            == types.MarkupKind.Markdown
+        ):
+            self.hover_output_style = MarkdownOutputStyle()
         self._finished_initialization = True
 
     async def on_initialize(self, params: types.InitializeParams) -> None:
@@ -470,6 +517,10 @@ class DebputyLanguageServer(LanguageServer):
             )
         _info(f"Cwd: {os.getcwd()}")
         self._update_locale()
+        self.workspace_text_edit_support = WorkspaceTextEditSupport(
+            supported_resource_operation_edit_kinds=self._supported_resource_operation_edit_kinds,
+            supports_document_changes=self._supports_ws_document_changes,
+        )
 
     def _update_locale(self) -> None:
         if self.forced_locale is not None:
@@ -479,6 +530,33 @@ class DebputyLanguageServer(LanguageServer):
         else:
             self._active_locale = None
 
+    def file_cache_for(
+        self,
+        uri: str,
+        path: str,
+        cache_type: Type[T],
+        initializer: Callable[[str, str], T],
+    ) -> T:
+        inner_cache = self._file_state_caches.get(path)
+        result = inner_cache.get(cache_type) if inner_cache else None
+        if result is None:
+            result = initializer(uri, path)
+            if uri not in self.workspace.text_documents:
+                # We do not get a proper notification on when to discard the cache.
+                # For now, we simply do not cache them at all to avoid "leaking by infinite cache".
+                #
+                # Note that even if we did cache them, we would have to track whether the files have
+                # changed (which we do not), so the cache itself would not be useful.
+                return result
+            assert isinstance(result, cache_type)
+            if inner_cache is None:
+                inner_cache = {}
+                self._file_state_caches[path] = inner_cache
+            inner_cache[cache_type] = result
+        else:
+            assert isinstance(result, cache_type)
+        return result
+
     def shutdown(self) -> None:
         for task in self.background_tasks:
             _info(f"Cancelling task: {task.get_name()}")
@@ -587,6 +665,26 @@ class DebputyLanguageServer(LanguageServer):
         except AttributeError:
             return None
 
+    @property
+    def _supports_ws_document_changes(self) -> bool:
+        try:
+            return (
+                self.client_capabilities.workspace.workspace_edit.document_changes  # type : ignore
+            )
+        except AttributeError:
+            return False
+
+    @property
+    def _supported_resource_operation_edit_kinds(
+        self,
+    ) -> Sequence[types.ResourceOperationKind]:
+        try:
+            return (
+                self.client_capabilities.workspace.workspace_edit.resource_operations  # type : ignore
+            )
+        except AttributeError:
+            return []
+
     def completion_item_document_markup(
         self,
         *options: types.MarkupKind,
@@ -638,10 +736,11 @@ class DebputyLanguageServer(LanguageServer):
         return "filename", cleaned_filename, cleaned_filename
 
     def close_document(self, uri: str) -> None:
-        try:
+        path = to_fs_path(uri)
+        with suppress(KeyError):
             del self._diagnostic_reports[uri]
-        except KeyError:
-            pass
+        with suppress(KeyError):
+            del self._file_state_caches[path]
 
     async def slow_iter(
         self,
