@@ -1,4 +1,5 @@
 import inspect
+import os.path
 from typing import (
     Sequence,
     Union,
@@ -88,7 +89,7 @@ async def _changed_document(
 
 
 @DEBPUTY_LANGUAGE_SERVER.feature(types.TEXT_DOCUMENT_DID_CLOSE)
-def _changed_document(
+def _close_document(
     ls: "DebputyLanguageServer",
     params: types.DidCloseTextDocumentParams,
 ) -> None:
@@ -149,7 +150,7 @@ async def _diagnostics_for(
         _trace_log(f"Aborted lint task: {e.message}")
         return
 
-    ls.record_diagnostics(doc_uri, expected_version, diagnostics, False)
+    ls.record_diagnostics(doc_uri, expected_version, diagnostics or [], False)
 
 
 @DEBPUTY_LANGUAGE_SERVER.feature(types.TEXT_DOCUMENT_COMPLETION)
@@ -229,7 +230,7 @@ async def _folding_ranges(
         full=True,
     ),
 )
-async def _semantic_tokens_full(
+async def semantic_tokens_full(
     ls: "DebputyLanguageServer",
     params: types.SemanticTokensParams,
 ) -> Optional[types.SemanticTokens]:
@@ -307,7 +308,18 @@ def _resolve_handler(
     log_func(f"resolve_handler({language_id=}, {normalized_filename=})")
     if dispatch_table is None:
         filename_based_table = handler_table[""]
-        return filename_based_table.filename_based_lookups.get(normalized_filename)
-    return dispatch_table.filename_based_lookups.get(
+        m = filename_based_table.basename_based_lookups.get(
+            os.path.basename(normalized_filename)
+        )
+        if m is not None:
+            return m
+        return filename_based_table.path_name_based_lookups.get(normalized_filename)
+    m = dispatch_table.basename_based_lookups.get(
+        os.path.basename(normalized_filename),
+        dispatch_table.default_handler,
+    )
+    if m is not None:
+        return m
+    return dispatch_table.path_name_based_lookups.get(
         normalized_filename, dispatch_table.default_handler
     )
