@@ -12,8 +12,46 @@ from typing import (
     Iterator,
     Callable,
     cast,
+    TYPE_CHECKING,
 )
 
+from debputy.linting.lint_util import (
+    LintState,
+    te_position_to_lsp,
+    with_range_in_continuous_parts,
+)
+from debputy.lsp.config.config_options import DCO_SPELLCHECK_COMMENTS
+from debputy.lsp.debputy_ls import DebputyLanguageServer
+from debputy.lsp.lsp_debian_control_reference_data import (
+    Deb822FileMetadata,
+    Deb822KnownField,
+    StanzaMetadata,
+    F,
+    S,
+    SUBSTVAR_RE,
+    _DEP_RELATION_CLAUSE,
+)
+from debputy.lsp.lsp_features import SEMANTIC_TOKEN_TYPES_IDS
+from debputy.lsp.quickfixes import propose_correct_text_quick_fix
+from debputy.lsp.text_util import (
+    trim_end_of_line_whitespace,
+    SemanticTokensState,
+)
+from debputy.lsp.vendoring._deb822_repro.locatable import (
+    START_POSITION,
+    Range as TERange,
+    Position as TEPosition,
+)
+from debputy.lsp.vendoring._deb822_repro.parsing import (
+    Deb822KeyValuePairElement,
+    Deb822ParagraphElement,
+    Deb822FileElement,
+    Deb822CommentElement,
+    Deb822ParsedTokenList,
+    Interpretation,
+)
+from debputy.lsp.vendoring._deb822_repro.tokens import tokenize_deb822_file, Deb822Token
+from debputy.lsp.vendoring._deb822_repro.types import TokenOrElement
 from debputy.lsprotocol.types import (
     CompletionParams,
     CompletionList,
@@ -32,34 +70,14 @@ from debputy.lsprotocol.types import (
     MessageType,
     SemanticTokenTypes,
 )
-
-from debputy.linting.lint_util import LintState, te_position_to_lsp
-from debputy.lsp.debputy_ls import DebputyLanguageServer
-from debputy.lsp.lsp_debian_control_reference_data import (
-    Deb822FileMetadata,
-    Deb822KnownField,
-    StanzaMetadata,
-    F,
-    S,
-)
-from debputy.lsp.lsp_features import SEMANTIC_TOKEN_TYPES_IDS
-from debputy.lsp.text_util import (
-    trim_end_of_line_whitespace,
-    SemanticTokensState,
-)
-from debputy.lsp.vendoring._deb822_repro.locatable import (
-    START_POSITION,
-    Range as TERange,
-)
-from debputy.lsp.vendoring._deb822_repro.parsing import (
-    Deb822KeyValuePairElement,
-    Deb822ParagraphElement,
-    Deb822FileElement,
-)
-from debputy.lsp.vendoring._deb822_repro.tokens import tokenize_deb822_file, Deb822Token
-from debputy.lsp.vendoring._deb822_repro.types import TokenOrElement
 from debputy.util import _info, _warn
 
+if TYPE_CHECKING:
+    import lsprotocol.types as types
+else:
+    import debputy.lsprotocol.types as types
+
+
 try:
     from pygls.server import LanguageServer
     from pygls.workspace import TextDocument
@@ -253,7 +271,7 @@ def deb822_completer(
         return None
 
     (
-        _a,
+        server_pos,
         current_field,
         word_at_position,
         in_value,
@@ -269,6 +287,9 @@ def deb822_completer(
         is_completion=True,
     )
 
+    if lines[server_pos.line].startswith("#"):
+        return
+
     items: Optional[Sequence[CompletionItem]]
     markdown_kind = ls.completion_item_document_markup(
         MarkupKind.Markdown, MarkupKind.PlainText
@@ -294,10 +315,6 @@ def deb822_completer(
             markdown_kind,
         )
 
-    _info(
-        f"Completion candidates: {[i.label for i in items] if items is not None else 'None'}"
-    )
-
     return items
 
 
@@ -347,6 +364,10 @@ def deb822_hover(
         lines,
         params.position,
     )
+
+    if lines[server_pos.line].startswith("#"):
+        return
+
     hover_text = None
     if custom_handler is not None:
         res = custom_handler(
@@ -465,6 +486,11 @@ class Deb822SemanticTokensState(SemanticTokensState):
         "keyword_token_code",
         "known_value_token_code",
         "comment_token_code",
+        "substvars_token_code",
+        "operator_token_code",
+        "relation_restriction_token_code",
+        "package_token_code",
+        "allow_overlapping_tokens",
     )
 
     def __init__(
@@ -474,20 +500,209 @@ class Deb822SemanticTokensState(SemanticTokensState):
         lines: List[str],
         tokens: List[int],
         file_metadata: Deb822FileMetadata[Any, Any],
-        keyword_token_code: int,
-        known_value_token_code: int,
-        comment_token_code: int,
     ) -> None:
         super().__init__(ls, doc, lines, tokens)
         self.file_metadata = file_metadata
-        self.keyword_token_code = keyword_token_code
-        self.known_value_token_code = known_value_token_code
-        self.comment_token_code = comment_token_code
+
+        self.keyword_token_code = SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Keyword]
+        self.known_value_token_code = SEMANTIC_TOKEN_TYPES_IDS[
+            SemanticTokenTypes.EnumMember
+        ]
+        self.comment_token_code = SEMANTIC_TOKEN_TYPES_IDS[
+            SemanticTokenTypes.Comment.value
+        ]
+        self.substvars_token_code = SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Macro]
+        self.operator_token_code = SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Operator]
+        self.relation_restriction_token_code = SEMANTIC_TOKEN_TYPES_IDS[
+            SemanticTokenTypes.TypeParameter
+        ]
+        self.package_token_code = SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Variable]
 
 
-def _deb822_paragraph_semantic_tokens_full(
+def _emit_tokens_for_comment_element(
+    sem_token_state: Deb822SemanticTokensState,
+    comment_element: Deb822CommentElement,
+    comment_start_line: int,
+    comment_token_code: int,
+) -> None:
+    for comment_line_no, comment_token in enumerate(
+        comment_element.iter_parts(),
+        start=comment_start_line,
+    ):
+        assert comment_token.is_comment
+        assert isinstance(comment_token, Deb822Token)
+        sem_token_state.emit_token(
+            Position(comment_line_no, 0),
+            len(comment_token.text.rstrip()),
+            comment_token_code,
+        )
+
+
+async def scan_for_syntax_errors_and_token_level_diagnostics(
+    deb822_file: Deb822FileElement,
+    lint_state: LintState,
+) -> int:
+    first_error = len(lint_state.lines) + 1
+    spell_checker = lint_state.spellchecker()
+
+    async for (
+        token,
+        start_line,
+        start_offset,
+        end_line,
+        end_offset,
+    ) in lint_state.slow_iter(deb822_token_iter(deb822_file.iter_tokens())):
+        if token.is_error:
+            first_error = min(first_error, start_line)
+            token_range = TERange(
+                TEPosition(
+                    start_line,
+                    start_offset,
+                ),
+                TEPosition(
+                    end_line,
+                    end_offset,
+                ),
+            )
+            lint_state.emit_diagnostic(
+                token_range,
+                "Syntax error",
+                "error",
+                "debputy",
+            )
+        elif token.is_comment:
+            if not lint_state.debputy_config.config_value(DCO_SPELLCHECK_COMMENTS):
+                continue
+
+            for word, col_pos, end_col_pos in spell_checker.iter_words(token.text):
+                corrections = spell_checker.provide_corrections_for(word)
+                if not corrections:
+                    continue
+                word_range = TERange.between(
+                    TEPosition(
+                        start_line,
+                        col_pos,
+                    ),
+                    TEPosition(
+                        start_line,
+                        end_col_pos,
+                    ),
+                )
+                lint_state.emit_diagnostic(
+                    word_range,
+                    f'Spelling "{word}"',
+                    "spelling",
+                    "debputy",
+                    quickfixes=[propose_correct_text_quick_fix(c) for c in corrections],
+                    enable_non_interactive_auto_fix=False,
+                )
+    return first_error
+
+
+def _emit_relation_token(
+    sem_token_state: Deb822SemanticTokensState,
+    token_code: Optional[int],
+    m: re.Match[str],
+    group_name: str,
+    value_range_te: TERange,
+) -> None:
+    token_value = m.group(group_name)
+    token_start = m.start(group_name)
+    if token_value is None or token_start is None:
+        return
+    pos = TEPosition(
+        value_range_te.start_pos.line_position,
+        value_range_te.start_pos.cursor_position + token_start,
+    )
+    end_pos = TEPosition(
+        pos.line_position,
+        pos.cursor_position + len(token_value),
+    )
+
+    _process_value_with_substvars(
+        sem_token_state,
+        token_value,
+        TERange.between(pos, end_pos),
+        token_code,
+    )
+
+
+async def _deb822_relationship_field_semantic_tokens_full(
+    sem_token_state: Deb822SemanticTokensState,
+    interpretation: Interpretation[Deb822ParsedTokenList[Any, Any]],
+    kvpair: Deb822KeyValuePairElement,
+    value_element_pos: TEPosition,
+) -> None:
+    doc = sem_token_state.doc
+    parts = kvpair.interpret_as(interpretation).iter_parts()
+    comment_token_code = sem_token_state.comment_token_code
+    operator_token_code = sem_token_state.operator_token_code
+    relation_restriction_token_code = sem_token_state.relation_restriction_token_code
+    package_token_code = sem_token_state.package_token_code
+
+    for te in parts:
+        if te.is_whitespace:
+            continue
+        if te.is_separator:
+            continue
+
+        value_range_in_parent_te = te.range_in_parent()
+        value_range_te = value_range_in_parent_te.relative_to(value_element_pos)
+        value = te.convert_to_text()
+        if te.is_comment:
+            token_type = comment_token_code
+            value = value.rstrip()
+            value_len = doc.position_codec.client_num_units(value)
+            sem_token_state.emit_token(
+                te_position_to_lsp(value_range_te.start_pos),
+                value_len,
+                token_type,
+            )
+        else:
+            m = _DEP_RELATION_CLAUSE.fullmatch(value)
+            _emit_relation_token(
+                sem_token_state,
+                package_token_code,
+                m,
+                "name_arch_qual",
+                value_range_te,
+            )
+
+            _emit_relation_token(
+                sem_token_state,
+                operator_token_code,
+                m,
+                "operator",
+                value_range_te,
+            )
+            _emit_relation_token(
+                sem_token_state,
+                None,
+                m,
+                "version",
+                value_range_te,
+            )
+            _emit_relation_token(
+                sem_token_state,
+                relation_restriction_token_code,
+                m,
+                "arch_restriction",
+                value_range_te,
+            )
+            _emit_relation_token(
+                sem_token_state,
+                relation_restriction_token_code,
+                m,
+                "build_profile_restriction",
+                value_range_te,
+            )
+
+
+async def _deb822_paragraph_semantic_tokens_full(
+    ls: "DebputyLanguageServer",
     sem_token_state: Deb822SemanticTokensState,
     stanza: Deb822ParagraphElement,
+    stanza_range_in_file: "TERange",
     stanza_idx: int,
 ) -> None:
     doc = sem_token_state.doc
@@ -495,30 +710,34 @@ def _deb822_paragraph_semantic_tokens_full(
     known_value_token_code = sem_token_state.known_value_token_code
     comment_token_code = sem_token_state.comment_token_code
 
-    stanza_position = stanza.position_in_file()
+    stanza_position = stanza_range_in_file.start_pos
     stanza_metadata = sem_token_state.file_metadata.classify_stanza(
         stanza,
         stanza_idx=stanza_idx,
     )
-    for kvpair in stanza.iter_parts_of_type(Deb822KeyValuePairElement):
-        kvpair_position = kvpair.position_in_parent().relative_to(stanza_position)
+    async for kvpair_range, kvpair in ls.slow_iter(
+        with_range_in_continuous_parts(
+            stanza.iter_parts(),
+            start_relative_to=stanza_position,
+        ),
+        yield_every=25,
+    ):
+        if not isinstance(kvpair, Deb822KeyValuePairElement):
+            continue
+        kvpair_position = kvpair_range.start_pos
         field_start = kvpair.field_token.position_in_parent().relative_to(
             kvpair_position
         )
         comment = kvpair.comment_element
         if comment:
             comment_start_line = field_start.line_position - len(comment)
-            for comment_line_no, comment_token in enumerate(
-                comment.iter_parts(),
-                start=comment_start_line,
-            ):
-                assert comment_token.is_comment
-                assert isinstance(comment_token, Deb822Token)
-                sem_token_state.emit_token(
-                    Position(comment_line_no, 0),
-                    len(comment_token.text.rstrip()),
-                    comment_token_code,
-                )
+            _emit_tokens_for_comment_element(
+                sem_token_state,
+                comment,
+                comment_start_line,
+                comment_token_code,
+            )
+
         field_size = doc.position_codec.client_num_units(kvpair.field_name)
 
         sem_token_state.emit_token(
@@ -528,20 +747,53 @@ def _deb822_paragraph_semantic_tokens_full(
         )
 
         known_field: Optional[Deb822KnownField] = stanza_metadata.get(kvpair.field_name)
-        if known_field is not None:
-            if known_field.spellcheck_value:
-                continue
-            known_values: Container[str] = known_field.known_values or frozenset()
-            interpretation = known_field.field_value_class.interpreter()
-        else:
-            known_values = frozenset()
-            interpretation = None
-
         value_element_pos = kvpair.value_element.position_in_parent().relative_to(
             kvpair_position
         )
+        if known_field is not None:
+            if known_field.spellcheck_value:
+                continue
+            interpretation = known_field.field_value_class.interpreter()
+            if (
+                getattr(known_field, "is_relationship_field", False)
+                and interpretation is not None
+            ):
+                await _deb822_relationship_field_semantic_tokens_full(
+                    sem_token_state,
+                    interpretation,
+                    kvpair,
+                    value_element_pos,
+                )
+                continue
+            known_values: Container[str] = known_field.known_values or frozenset()
+            field_disallows_substvars = (
+                known_field.is_substvars_disabled_even_if_allowed_by_stanza
+            )
+            allow_substvars = (
+                stanza_metadata.is_substvars_allowed_in_stanza
+                and not field_disallows_substvars
+            )
+        else:
+            known_values = frozenset()
+            interpretation = None
+            allow_substvars = stanza_metadata.is_substvars_allowed_in_stanza
+
         if interpretation is None:
-            # TODO: Emit tokens for value comments of unknown fields.
+            for value_line in kvpair.value_element.value_lines:
+                comment_element = value_line.comment_element
+                if comment_element:
+                    assert comment_element.position_in_parent().line_position == 0
+                    comment_start_line = (
+                        value_line.position_in_parent()
+                        .relative_to(value_element_pos)
+                        .line_position
+                    )
+                    _emit_tokens_for_comment_element(
+                        sem_token_state,
+                        comment_element,
+                        comment_start_line,
+                        comment_token_code,
+                    )
             continue
         else:
             parts = kvpair.interpret_as(interpretation).iter_parts()
@@ -558,10 +810,17 @@ def _deb822_paragraph_semantic_tokens_full(
                 value = value.rstrip()
             elif value in known_values:
                 token_type = known_value_token_code
+            elif allow_substvars and "${" in value:
+                _process_value_with_substvars(
+                    sem_token_state,
+                    value,
+                    value_range_te,
+                    None,
+                )
+                continue
             else:
                 continue
             value_len = doc.position_codec.client_num_units(value)
-
             sem_token_state.emit_token(
                 te_position_to_lsp(value_range_te.start_pos),
                 value_len,
@@ -569,6 +828,70 @@ def _deb822_paragraph_semantic_tokens_full(
             )
 
 
+def _split_into_substvars(
+    value: str,
+    base_token_type: Optional[int],
+    substvar_token_type: int,
+) -> Iterable[Tuple[str, Optional[int]]]:
+
+    i = 0
+    next_search = i
+    full_value_len = len(value)
+    while i < full_value_len:
+        try:
+            subst_var_start = value.index("${", next_search)
+            subst_var_end = value.index("}", subst_var_start + 2)
+        except ValueError:
+            token = value[i:full_value_len]
+            if token:
+                yield token, base_token_type
+            return
+
+        subst_var_end += 1
+        subst_var = value[subst_var_start:subst_var_end]
+        if subst_var != "${}" and not SUBSTVAR_RE.match(subst_var):
+            subst_var = None
+
+        if subst_var is None:
+            next_search = subst_var_end
+            continue
+
+        token = value[i:subst_var_start]
+        if token:
+            yield token, base_token_type
+        yield subst_var, substvar_token_type
+        i = subst_var_end
+        next_search = i
+
+
+def _process_value_with_substvars(
+    sem_token_state: Deb822SemanticTokensState,
+    value: str,
+    value_range_te: "TERange",
+    base_token_type: Optional[int],
+) -> None:
+    pos_codec = sem_token_state.doc.position_codec
+
+    # TODO: Support overlapping tokens if the editor does.
+
+    line = value_range_te.start_pos.line_position
+    token_pos = value_range_te.start_pos.cursor_position
+    substvar_token_code = sem_token_state.substvars_token_code
+    for token, token_type in _split_into_substvars(
+        value,
+        base_token_type,
+        substvar_token_code,
+    ):
+        token_len = len(token)
+        if token_type is not None:
+            sem_token_state.emit_token(
+                types.Position(line, token_pos),
+                pos_codec.client_num_units(token),
+                token_type,
+            )
+        token_pos += token_len
+
+
 def deb822_format_file(
     lint_state: LintState,
     file_metadata: Deb822FileMetadata[Any, Any],
@@ -595,7 +918,7 @@ def deb822_format_file(
     )
 
 
-def deb822_semantic_tokens_full(
+async def deb822_semantic_tokens_full(
     ls: "DebputyLanguageServer",
     request: SemanticTokensParams,
     file_metadata: Deb822FileMetadata[Any, Any],
@@ -612,23 +935,23 @@ def deb822_semantic_tokens_full(
         return None
 
     tokens: List[int] = []
-    comment_token_code = SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Comment.value]
     sem_token_state = Deb822SemanticTokensState(
         ls,
         doc,
         lines,
         tokens,
         file_metadata,
-        SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Keyword],
-        SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.EnumMember],
-        comment_token_code,
     )
 
+    comment_token_code = sem_token_state.comment_token_code
+
     stanza_idx = 0
 
-    for part in deb822_file.iter_parts():
+    async for part_range, part in ls.slow_iter(
+        with_range_in_continuous_parts(deb822_file.iter_parts()), yield_every=20
+    ):
         if part.is_comment:
-            pos = part.position_in_file()
+            pos = part_range.start_pos
             sem_token_state.emit_token(
                 te_position_to_lsp(pos),
                 # Avoid trailing newline
@@ -636,9 +959,11 @@ def deb822_semantic_tokens_full(
                 comment_token_code,
             )
         elif isinstance(part, Deb822ParagraphElement):
-            _deb822_paragraph_semantic_tokens_full(
+            await _deb822_paragraph_semantic_tokens_full(
+                ls,
                 sem_token_state,
                 part,
+                part_range,
                 stanza_idx,
             )
             stanza_idx += 1
