@@ -9,8 +9,10 @@ from typing import (
     Iterable,
     Literal,
     Dict,
+    Union,
 )
 
+from debputy.lsp.diagnostics import LintSeverity
 from debputy.lsp.vendoring._deb822_repro import (
     LIST_SPACE_SEPARATED_INTERPRETATION,
     LIST_COMMA_SEPARATED_INTERPRETATION,
@@ -138,8 +140,11 @@ class Deb822Field(DebputyParsedContent):
     canonical_name: str
     field_value_type: FieldValueType
     unknown_value_authority: NotRequired[str]
+    unknown_value_severity: NotRequired[Union[LintSeverity, Literal["none"]]]
     missing_field_authority: NotRequired[str]
+    missing_field_severity: NotRequired[LintSeverity]
     default_value: NotRequired[str]
+    warn_if_default: NotRequired[bool]
     usage_hint: NotRequired[UsageHint]
     documentation: NotRequired[Documentation]
     values: NotRequired[List[StaticValue]]
